@@ -1,8 +1,9 @@
 package Dancer2::Core::Request;
-$Dancer2::Core::Request::VERSION = '0.158000';
 # ABSTRACT: Interface for accessing incoming requests
-
-use Moo;
+$Dancer2::Core::Request::VERSION = '0.166001';
+use strict;
+use warnings;
+use parent 'Plack::Request';
 
 use Carp;
 use Encode;
@@ -10,18 +11,16 @@ use HTTP::Body;
 use URI;
 use URI::Escape;
 use Class::Load 'try_load_class';
-use Plack::Request;
+use Safe::Isa;
+use Hash::MultiValue;
 
 use Dancer2::Core::Types;
 use Dancer2::Core::Request::Upload;
 use Dancer2::Core::Cookie;
 
-with 'Dancer2::Core::Role::Headers';
-
 # add an attribute for each HTTP_* variables
 # (HOST is managed manually)
 my @http_env_keys = (qw/
-    accept
     accept_charset
     accept_encoding
     accept_language
@@ -30,41 +29,63 @@ my @http_env_keys = (qw/
     x_requested_with
 /);
 
-foreach my $attr ( @http_env_keys ) {
-    has $attr => (
-        is      => 'ro',
-        isa     => Maybe[Str],
-        lazy    => 1,
-        default => sub { $_[0]->env->{ 'HTTP_' . ( uc $attr ) } },
-    );
-}
+# apparently you can't eval core functions
+sub accept { $_[0]->env->{'HTTP_ACCEPT'} }
+
+eval << "_EVAL" for @http_env_keys; ## no critic
+sub $_ { \$_[0]->env->{ 'HTTP_' . ( uc $_ ) } }
+_EVAL
 
 # check presence of XS module to speedup request
 our $XS_URL_DECODE         = try_load_class('URL::Encode::XS');
 our $XS_PARSE_QUERY_STRING = try_load_class('CGI::Deurl::XS');
 
-our $_count = 1;
+our $_id = 0;
 
-# then all the native attributes
-has env => (
-    is       => 'ro',
-    isa      => HashRef,
-    required => 1,
-);
+# self->new( env => {}, serializer => $s, is_behind_proxy => 0|1 )
+sub new {
+    my ( $class, @args ) = @_;
 
-has _preq => (
-    is      => 'ro',
-    isa     => InstanceOf['Plack::Request'],
-    lazy    => 1, # make sure it's after env has been verified
-    default => sub { Plack::Request->new( $_[0]->env ) },
-);
+    # even sized list
+    @args % 2 == 0
+        or croak 'Must provide even sized list';
+
+    my %opts = @args;
+    my $env  = $opts{'env'};
+
+    my $self = $class->SUPER::new($env);
+
+    if ( my $s = $opts{'serializer'} ) {
+        $s->$_does('Dancer2::Core::Role::Serializer')
+            or croak 'Serializer provided not a Serializer object';
+
+        $self->{'serializer'} = $s;
+    }
+
+    # additionally supported attributes
+    $self->{'id'}              = ++$_id;
+    $self->{'vars'}            = {};
+    $self->{'is_behind_proxy'} = !!$opts{'is_behind_proxy'};
+
+    # parameters
+    $self->{_chunk_size}       = 4096;
+    $self->{_read_position}    = 0;
+    $self->{_http_body} =
+      HTTP::Body->new( $self->content_type || '', $self->content_length );
+    $self->{_http_body}->cleanup(1);
+
+    $opts{'body_params'}
+        and $self->{'_body_params'} = $opts{'body_params'};
+
+    # parsing body for HMV first (also deserializes body)
+    $self->body_parameters;
+    $self->_build_uploads();
+
+    return $self;
+}
 
 # a buffer for per-request variables
-has vars => (
-    is      => 'ro',
-    isa     => HashRef,
-    default => sub { {} },
-);
+sub vars { $_[0]->{'vars'} }
 
 sub var {
     my $self = shift;
@@ -73,32 +94,15 @@ sub var {
       : $self->vars->{ $_[0] };
 }
 
-has path => (
-    is      => 'ro',
-    isa     => Str,
-    lazy    => 1,
-    default => sub { $_[0]->path_info || '/' },
-);
+# I don't like this. I know send_file uses this and I wonder
+# if we can remove it.
+#   -- Sawyer
+sub set_path_info { $_[0]->env->{'PATH_INFO'} = $_[1] }
 
-has path_info => (
-    is      => 'ro',
-    isa     => Str,
-    lazy    => 1,
-    writer  => 'set_path_info',
-    default => sub { $_[0]->env->{'PATH_INFO'} },
-);
+# XXX: incompatible with Plack::Request
+sub body { $_[0]->{'body'} ||= $_[0]->_read_to_end }
 
-has body => (
-    is      => 'ro',
-    isa     => Str,
-    default => sub {''},
-);
-
-has id => (
-    is      => 'ro',
-    isa     => Num,
-    default => sub { $_count++ },
-);
+sub id { $_id }
 
 # Private 'read-only' attributes for request params. See the params()
 # method for the public interface.
@@ -107,43 +111,27 @@ has id => (
 # decode byte string to characters before setting; If you know you have
 # decoded (character) params, such as output from a deserializer, you can
 # set these directly in the request object hash to avoid the decode op.
+sub _params { $_[0]->{'_params'} ||= $_[0]->_build_params }
 
-has _params => (
-    is        => 'lazy',
-    isa       => HashRef,
-    builder   => '_build_params',
-    predicate => '_has_params',
-);
+sub _has_params { defined $_[0]->{'_params'} }
 
-has _body_params => (
-    is      => 'ro',
-    isa     => Maybe[HashRef],
-    default => sub {undef},
-);
+sub _body_params {
+    my $self = shift;
 
-sub _set_body_params {
-    my ( $self, $params ) = @_;
-    $self->{_body_params} = _decode( $params );
-    $self->_build_params();
+    # make sure body is parsed
+    $self->body;
+
+    $self->{'_body_params'} ||= _decode( $self->{'_http_body'}->param );
 }
 
-has _query_params => (
-    is      => 'ro',
-    isa     => Maybe[HashRef],
-    default => sub {undef},
-);
+sub _query_params { $_[0]->{'_query_params'} }
 
 sub _set_query_params {
     my ( $self, $params ) = @_;
     $self->{_query_params} = _decode( $params );
-    $self->_build_params();
 }
 
-has _route_params => (
-    is      => 'ro',
-    isa     => HashRef,
-    default => sub {{}},
-);
+sub _route_params { $_[0]->{'_route_params'} ||= {} }
 
 sub _set_route_params {
     my ( $self, $params ) = @_;
@@ -151,45 +139,16 @@ sub _set_route_params {
     $self->_build_params();
 }
 
-has uploads => (
-    is  => 'ro',
-    isa => HashRef,
-);
+# XXX: incompatible with Plack::Request
+sub uploads { $_[0]->{'uploads'} }
 
-has body_is_parsed => (
-    is      => 'ro',
-    isa     => Bool,
-    default => sub {0},
-);
-
-has is_behind_proxy => (
-    is      => 'ro',
-    isa     => Bool,
-    lazy    => 1,
-    default => sub {0},
-);
-
-sub BUILDARGS {
-    my ( $class, @args ) = @_;
-
-    # regular hash (or empty)
-    @args != 1 and return {@args};
-
-    # we don't know what this is
-    ref $args[0] eq 'HASH' or return {@args};
-
-    # handle what could be a single arg env
-    $args[0]->{'env'} or return { env => $args[0] };
-
-    # everything else
-    return $args[0];
-}
+sub is_behind_proxy { $_[0]->{'is_behind_proxy'} || 0 }
 
 sub host {
     my ($self) = @_;
 
-    if ( $self->is_behind_proxy ) {
-        my @hosts = split /\s*,\s*/, $self->env->{HTTP_X_FORWARDED_HOST}, 2;
+    if ( $self->is_behind_proxy and exists $self->env->{'HTTP_X_FORWARDED_HOST'} ) {
+        my @hosts = split /\s*,\s*/, $self->env->{'HTTP_X_FORWARDED_HOST'}, 2;
         return $hosts[0];
     } else {
         return $self->env->{'HTTP_HOST'};
@@ -202,61 +161,31 @@ sub remote_address        { shift->address }
 sub forwarded_for_address { shift->env->{'HTTP_X_FORWARDED_FOR'} }
 sub forwarded_host        { shift->env->{'HTTP_X_FORWARDED_HOST'} }
 
-# attributes
-sub address               { $_[0]->_preq->address }
-sub remote_host           { $_[0]->_preq->remote_host }
-sub protocol              { $_[0]->_preq->protocol }
-sub port                  { $_[0]->_preq->port }
-sub method                { $_[0]->_preq->method }
-sub user                  { $_[0]->_preq->user }
-sub request_uri           { $_[0]->_preq->request_uri }
-sub script_name           { $_[0]->_preq->script_name }
-sub content_length        { $_[0]->_preq->content_length }
-sub content_type          { $_[0]->_preq->content_type }
-sub secure                { $_[0]->scheme eq 'https' }
-
-# headers
-sub content_encoding      { $_[0]->_preq->content_encoding }
-sub referer               { $_[0]->_preq->referer }
-sub user_agent            { $_[0]->_preq->user_agent }
-
 # there are two options
 sub forwarded_protocol    {
-    $_[0]->env->{HTTP_X_FORWARDED_PROTO} ||
-    $_[0]->env->{HTTP_X_FORWARDED_PROTOCOL}
+    $_[0]->env->{'HTTP_X_FORWARDED_PROTOCOL'} ||
+    $_[0]->env->{'HTTP_X_FORWARDED_PROTO'}    ||
+    $_[0]->env->{'HTTP_FORWARDED_PROTO'}
 }
 
 sub scheme {
     my ($self) = @_;
-    my $scheme;
-    if ( $self->is_behind_proxy ) {
-        # Note the 'HTTP_' prefix the PSGI spec adds to headers.
-        $scheme =
-             $self->env->{'HTTP_X_FORWARDED_PROTOCOL'}
-          || $self->env->{'HTTP_X_FORWARDED_PROTO'}
-          || $self->env->{'HTTP_FORWARDED_PROTO'}
-          || '';
-    }
+    my $scheme = $self->is_behind_proxy
+               ? $self->forwarded_protocol
+               : '';
 
-    return $scheme || $self->env->{'psgi.url_scheme'} || '';
+    return $scheme || $self->env->{'psgi.url_scheme'};
 }
 
-has serializer => (
-    is        => 'ro',
-    isa       => Maybe[ ConsumerOf['Dancer2::Core::Role::Serializer'] ],
-    predicate => 1,
-);
+sub serializer { $_[0]->{'serializer'} }
 
-has data => (
-    is      => 'ro',
-    lazy    => 1,
-    default => \&deserialize,
-);
+sub data { $_[0]->{'data'} ||= &deserialize }
 
 sub deserialize {
     my $self = shift;
 
-    return unless $self->has_serializer;
+    my $serializer = $self->serializer
+        or return;
 
     # The latest draft of the RFC does not forbid DELETE to have content,
     # rather the behaviour is undefined. Take the most lenient route and
@@ -265,12 +194,12 @@ sub deserialize {
       unless grep { $self->method eq $_ } qw/ PUT POST PATCH DELETE /;
 
     # try to deserialize
-    my $body = $self->_read_to_end();
+    my $body = $self->body;
 
     $body && length $body > 0
         or return;
 
-    my $data = $self->serializer->deserialize($self->body);
+    my $data = $serializer->deserialize($body);
     return if !defined $data;
 
     # Set _body_params directly rather than using the setter. Deserializiation
@@ -278,40 +207,24 @@ sub deserialize {
     # that numerical data "stays" numerical; decoding an SV that is an IV
     # converts that to a PVIV. Some serializers are picky (JSON)..
     $self->{_body_params} = $data;
-    $self->_build_params();
 
     return $data;
 }
 
-sub uri       { $_[0]->request_uri }
-sub is_head   { $_[0]->method eq 'HEAD' }
-sub is_post   { $_[0]->method eq 'POST' }
-sub is_get    { $_[0]->method eq 'GET' }
-sub is_put    { $_[0]->method eq 'PUT' }
-sub is_delete { $_[0]->method eq 'DELETE' }
-sub is_patch  { $_[0]->method eq 'PATCH' }
+sub uri        { $_[0]->request_uri }
+
+sub is_head    { $_[0]->method eq 'HEAD' }
+sub is_post    { $_[0]->method eq 'POST' }
+sub is_get     { $_[0]->method eq 'GET' }
+sub is_put     { $_[0]->method eq 'PUT' }
+sub is_delete  { $_[0]->method eq 'DELETE' }
+sub is_patch   { $_[0]->method eq 'PATCH' }
+sub is_options { $_[0]->method eq 'OPTIONS' }
 
 # public interface compat with CGI.pm objects
-sub request_method { method(@_) }
+sub request_method { $_[0]->method }
 sub input_handle { $_[0]->env->{'psgi.input'} }
 
-sub BUILD {
-    my ($self) = @_;
-
-    $self->{_chunk_size}    = 4096;
-    $self->{_read_position} = 0;
-
-    $self->_init_request_headers();
-
-    $self->{_http_body} =
-      HTTP::Body->new( $self->content_type || '', $self->content_length );
-    $self->{_http_body}->cleanup(1);
-
-    $self->data;      # Deserialize body
-    $self->_params(); # Decode query and body prams
-    $self->_build_uploads();
-}
-
 sub to_string {
     my ($self) = @_;
     return "[#" . $self->id . "] " . $self->method . " " . $self->path;
@@ -414,10 +327,78 @@ sub params {
     }
 }
 
-sub captures { shift->params->{captures} }
+sub query_parameters {
+    my $self = shift;
+    $self->{'query_parameters'} ||= do {
+        if ($XS_PARSE_QUERY_STRING) {
+            my $query = CGI::Deurl::XS::parse_query_string(
+                $self->env->{'QUERY_STRING'}
+            );
+
+            Hash::MultiValue->new(
+                map {;
+                    my $key = $_;
+                    ref $query->{$key} eq 'ARRAY'
+                    ? ( map +( $key => $_ ), @{ $query->{$key} } )
+                    : ( $key => $query->{$key} )
+                } keys %{$query}
+            );
+        } else {
+            # defer to Plack::Request
+            $self->_parse_query;
+        }
+    };
+}
+
+# this will be filled once the route is matched
+sub route_parameters { $_[0]->{'route_parameters'} ||= Hash::MultiValue->new }
+
+sub _set_route_parameters {
+    my ( $self, $params ) = @_;
+    # remove reserved splat parameter name
+    # you should access splat parameters using splat() keyword
+    delete @{$params}{qw<splat captures>};
+    $self->{'route_parameters'} = Hash::MultiValue->from_mixed( %{$params} );
+}
+
+sub body_parameters {
+    my $self = shift;
+    $self->{'plack.request.body'}
+        and return $self->{'plack.request.body'};
+
+    # handle case of serializer
+    if ( my $data = $self->deserialize ) {
+        return Hash::MultiValue->from_mixed(
+            ref $data eq 'HASH' ? %{$data} : ()
+        );
+    }
+
+    $self->_parse_request_body;
+    return $self->env->{'plack.request.body'};
+}
+
+sub parameters {
+    my ( $self, $type ) = @_;
+
+    # handle a specific case
+    if ($type) {
+        my $attr = "${type}_parameters";
+        return $self->$attr;
+    }
+
+    $self->env->{'plack.request.merged'} ||= do {
+        my $query = $self->query_parameters;
+        my $body  = $self->body_parameters;
+        my $route = $self->route_parameters; # not in Plack::Request
+        Hash::MultiValue->new( map $_->flatten, $query, $body, $route );
+    };
+}
+
+sub captures { shift->params->{captures} || {} }
 
 sub splat { @{ shift->params->{splat} || [] } }
 
+# XXX: incompatible with Plack::Request
 sub param { shift->params->{ $_[0] } }
 
 sub _decode {
@@ -429,14 +410,11 @@ sub _decode {
     }
 
     if ( ref($h) eq 'HASH' ) {
-        while ( my ( $k, $v ) = each(%$h) ) {
-            $h->{$k} = _decode($v);
-        }
-        return $h;
+        return { map {my $t = _decode($_); $t} (%$h) };
     }
 
     if ( ref($h) eq 'ARRAY' ) {
-        return [ map { _decode($_) } @$h ];
+        return [ map _decode($_), @$h ];
     }
 
     return $h;
@@ -451,6 +429,7 @@ sub is_ajax {
     return 1;
 }
 
+# XXX incompatible with Plack::Request
 # context-aware accessor for uploads
 sub upload {
     my ( $self, $name ) = @_;
@@ -469,18 +448,15 @@ sub _build_params {
     my $previous = $self->_has_params ? $self->_params : {};
 
     # now parse environment params...
-    $self->_parse_get_params();
-    if ( $self->body_is_parsed ) {
-        $self->{_body_params} ||= {};
-    }
-    else {
-        $self->_parse_post_params();
-    }
+    my $get_params = $self->_parse_get_params();
 
     # and merge everything
     $self->{_params} = {
-        %$previous,                %{ $self->_query_params || {} },
-        %{ $self->_route_params }, %{ $self->_body_params  || {} },
+        map +( ref $_ eq 'HASH' ? %{$_} : () ),
+        $previous,
+        $get_params,
+        $self->_body_params,
+        $self->_route_params,
     };
 
 }
@@ -494,14 +470,6 @@ sub _url_decode {
     return $clean;
 }
 
-sub _parse_post_params {
-    my ($self) = @_;
-    return $self->_body_params if defined $self->_body_params;
-
-    my $body = $self->_read_to_end();
-    $self->_set_body_params( $self->{_http_body}->param );
-}
-
 sub _parse_get_params {
     my ($self) = @_;
     return $self->_query_params if defined $self->{_query_params};
@@ -551,14 +519,15 @@ sub _read_to_end {
     my $content_length = $self->content_length;
     return unless $self->_has_something_to_read();
 
+    my $body = '';
     if ( $content_length && $content_length > 0 ) {
         while ( my $buffer = $self->_read() ) {
-            $self->{body} .= $buffer;
+            $body .= $buffer;
         }
-        $self->{_http_body}->add( $self->{body} );
+        $self->{_http_body}->add($body);
     }
 
-    return $self->{body};
+    return $body;
 }
 
 sub _has_something_to_read {
@@ -568,7 +537,7 @@ sub _has_something_to_read {
 
 # taken from Miyagawa's Plack::Request::BodyParser
 sub _read {
-    my ( $self, ) = @_;
+    my ( $self ) = @_;
     my $remaining = $self->content_length - $self->{_read_position};
     my $maxlength = $self->{_chunk_size};
 
@@ -589,25 +558,13 @@ sub _read {
     }
 }
 
-sub _init_request_headers {
-    my ($self) = @_;
-    my $env = $self->env;
-
-    $self->headers(
-        HTTP::Headers->new(
-            map {
-                ( my $field = $_ ) =~ s/^HTTPS?_//;
-                ( $field => $env->{$_} );
-              }
-              grep {/^(?:HTTP|CONTENT)/i} keys %$env
-        )
-    );
-}
-
 # Taken gently from Plack::Request, thanks to Plack authors.
 sub _build_uploads {
     my ($self) = @_;
 
+    # build the body and body params
+    my $body_params = $self->_body_params;
+
     my $uploads = _decode( $self->{_http_body}->upload );
     my %uploads;
 
@@ -615,36 +572,26 @@ sub _build_uploads {
         my $files = $uploads->{$name};
         $files = ref $files eq 'ARRAY' ? $files : [$files];
 
-        my @uploads;
-        for my $upload ( @{$files} ) {
-            push(
-                @uploads,
-                Dancer2::Core::Request::Upload->new(
-                    headers  => $upload->{headers},
-                    tempname => $upload->{tempname},
-                    size     => $upload->{size},
-                    filename => $upload->{filename},
-                )
-            );
-        }
+        my @uploads = map Dancer2::Core::Request::Upload->new(
+                              headers  => $_->{headers},
+                              tempname => $_->{tempname},
+                              size     => $_->{size},
+                              filename => $_->{filename},
+                      ), @{$files};
+
         $uploads{$name} = @uploads > 1 ? \@uploads : $uploads[0];
 
         # support access to the filename as a normal param
-        my @filenames = map { $_->{filename} } @uploads;
+        my @filenames = map $_->{'filename'}, @uploads;
         $self->{_body_params}{$name} =
           @filenames > 1 ? \@filenames : $filenames[0];
     }
 
     $self->{uploads} = \%uploads;
-    $self->_build_params();
 }
 
-has cookies => (
-    is      => 'ro',
-    isa     => HashRef,
-    lazy    => 1,
-    builder => '_build_cookies',
-);
+# XXX: incompatible with Plack::Request
+sub cookies { $_[0]->{'cookies'} ||= $_[0]->_build_cookies }
 
 sub _build_cookies {
     my $self    = shift;
@@ -661,8 +608,8 @@ sub _build_cookies {
         # we want `cookie_name' as the value and `foo=bar' as the value
         my ( $name, $value ) = split( /\s*=\s*/, $cookie, 2 );
         my @values;
-        if ( $value ne '' ) {
-            @values = map { uri_unescape($_) } split( /[&;]/, $value );
+        if ( defined $value and $value ne '' ) {
+            @values = map uri_unescape($_), split( /[&;]/, $value );
         }
         $cookies->{$name} =
           Dancer2::Core::Cookie->new( name => $name, value => \@values );
@@ -684,34 +631,326 @@ Dancer2::Core::Request - Interface for accessing incoming requests
 
 =head1 VERSION
 
-version 0.158000
+version 0.166001
 
 =head1 SYNOPSIS
 
-In a route handler, the current request object can be accessed by the C<request>
-method, like in the following example:
+In a route handler, the current request object can be accessed by the
+C<request> keyword:
 
     get '/foo' => sub {
         request->params; # request, params parsed as a hash ref
-        request->body; # returns the request body, unparsed
-        request->path; # the path requested by the client
+        request->body;   # returns the request body, unparsed
+        request->path;   # the path requested by the client
         # ...
     };
 
-A route handler should not read the environment by itself, but should instead
-use the current request object.
-
 =head1 DESCRIPTION
 
-This class implements a common interface for accessing incoming requests in
-a Dancer2 application.
+An object representing a Dancer2 request. It aims to provide a proper
+interface to anything you might need from a web request.
 
 =head1 METHODS
 
-=head2 env()
+=head2 address
+
+Return the IP address of the client.
+
+=head2 base
+
+Returns an absolute URI for the base of the application.  Returns a L<URI>
+object (which stringifies to the URL, as you'd expect).
+
+=head2 body_parameters
+
+Returns a L<Hash::MultiValue> object representing the POST parameters.
+
+=head2 body
+
+Return the raw body of the request, unparsed.
+
+If you need to access the body of the request, you have to use this accessor and
+should not try to read C<psgi.input> by hand. C<Dancer2::Core::Request>
+already did it for you and kept the raw body untouched in there.
+
+=head2 content
+
+Returns the undecoded byte string POST body.
+
+=head2 cookies
+
+Returns a reference to a hash containing cookies, where the keys are the names of the
+cookies and values are L<Dancer2::Core::Cookie> objects.
+
+=head2 data
+
+If the application has a serializer and if the request has serialized
+content, returns the deserialized structure as a hashref.
+
+=head2 dispatch_path
+
+The part of the C<path> after C<base>. This is the path used
+for dispatching the request to routes.
+
+=head2 env
 
 Return the current PSGI environment hash reference.
 
+=head2 header($name)
+
+Return the value of the given header, if present. If the header has multiple
+values, returns an the list of values if called in list context, the first one
+in scalar.
+
+=head2 headers
+
+Returns either an L<HTTP::Headers> or an L<HTTP::Headers::Fast> object
+representing the headers.
+
+=head2 id
+
+The ID of the request. This allows you to trace a specific request in loggers,
+per the string created using C<to_string>.
+
+The ID of the request is essentially the number of requests run in the current
+class.
+
+=head2 input
+
+Alias to C<input_handle> method below.
+
+=head2 input_handle
+
+Alias to the PSGI input handle (C<< <request->env->{psgi.input}> >>)
+
+=head2 is_ajax
+
+Return true if the value of the header C<X-Requested-With> is
+C<XMLHttpRequest>.
+
+=head2 is_delete
+
+Return true if the method requested by the client is 'DELETE'
+
+=head2 is_get
+
+Return true if the method requested by the client is 'GET'
+
+=head2 is_head
+
+Return true if the method requested by the client is 'HEAD'
+
+=head2 is_post
+
+Return true if the method requested by the client is 'POST'
+
+=head2 is_put
+
+Return true if the method requested by the client is 'PUT'
+
+=head2 is_options
+
+Return true if the method requested by the client is 'OPTIONS'
+
+=head2 logger
+
+Returns the C<psgix.logger> code reference, if exists.
+
+=head2 method
+
+Return the HTTP method used by the client to access the application.
+
+While this method returns the method string as provided by the environment, it's
+better to use one of the following boolean accessors if you want to inspect the
+requested method.
+
+=head2 new
+
+The constructor of the class, used internally by Dancer2's core to create request
+objects.
+
+It uses the environment hash table given to build the request object:
+
+    Dancer2::Core::Request->new( env => $env );
+
+There are two additional parameters for instantiation:
+
+=over 4
+
+=item * serializer
+
+A serializer object to work with when reading the request body.
+
+=item * body_params
+
+Provide body parameters.
+
+Used internally when we need to avoid parsing the body again.
+
+=back
+
+=head2 param($key)
+
+Calls the C<params> method below and fetches the key provided.
+
+=head2 params($source)
+
+Called in scalar context, returns a hashref of params, either from the specified
+source (see below for more info on that) or merging all sources.
+
+So, you can use, for instance:
+
+    my $foo = params->{foo}
+
+If called in list context, returns a list of key and value pairs, so you could use:
+
+    my %allparams = params;
+
+=head2 parameters
+
+Returns a L<Hash::MultiValue> object with merged GET and POST parameters.
+
+=head2 path
+
+The path requested by the client, normalized. This is effectively
+C<path_info> or a single forward C</>.
+
+=head2 path_info
+
+The raw requested path. This could be empty. Use C<path> instead.
+
+=head2 port
+
+Return the port of the server.
+
+=head2 protocol
+
+Return the protocol (I<HTTP/1.0> or I<HTTP/1.1>) used for the request.
+
+=head2 query_parameters
+
+Returns a L<Hash::MultiValue> parameters object.
+
+=head2 query_string
+
+Returns the portion of the request defining the query itself - this is
+what comes after the C<?> in a URI.
+
+=head2 raw_body
+
+Alias to C<content> method.
+
+=head2 remote_address
+
+Alias for C<address> method.
+
+=head2 remote_host
+
+Return the remote host of the client. This only works with web servers configured
+to do a reverse DNS lookup on the client's IP address.
+
+=head2 request_method
+
+Alias to the C<method> accessor, for backward-compatibility with C<CGI> interface.
+
+=head2 request_uri
+
+Return the raw, undecoded request URI path.
+
+=head2 scheme
+
+Return the scheme of the request
+
+=head2 script_name
+
+Return script_name from the environment.
+
+=head2 secure
+
+Return true or false, indicating whether the connection is secure - this is
+effectively checking if the scheme is I<HTTPS> or not.
+
+=head2 serializer
+
+Returns the optional serializer object used to deserialize request parameters.
+
+=head2 session
+
+Returns the C<psgix.session> hash, if exists.
+
+=head2 session_options
+
+Returns the C<psgix.session.options> hash, if exists.
+
+=head2 to_string
+
+Return a string representing the request object (e.g., C<GET /some/path>).
+
+=head2 upload($name)
+
+Context-aware accessor for uploads. It's a wrapper around an access to the hash
+table provided by C<uploads()>. It looks at the calling context and returns a
+corresponding value.
+
+If you have many file uploads under the same name, and call C<upload('name')> in
+an array context, the accessor will unroll the ARRAY ref for you:
+
+    my @uploads = request->upload('many_uploads'); # OK
+
+Whereas with a manual access to the hash table, you'll end up with one element
+in C<@uploads>, being the arrayref:
+
+    my @uploads = request->uploads->{'many_uploads'};
+    # $uploads[0]: ARRAY(0xXXXXX)
+
+That is why this accessor should be used instead of a manual access to
+C<uploads>.
+
+=head2 uploads
+
+Returns a reference to a hash containing uploads. Values can be either a
+L<Dancer2::Core::Request::Upload> object, or an arrayref of
+L<Dancer2::Core::Request::Upload>
+objects.
+
+You should probably use the C<upload($name)> accessor instead of manually accessing the
+C<uploads> hash table.
+
+=head2 uri
+
+An alias to C<request_uri>.
+
+=head2 uri_base
+
+Same thing as C<base> above, except it removes the last trailing slash in the
+path if it is the only path.
+
+This means that if your base is I<http://myserver/>, C<uri_base> will return
+I<http://myserver> (notice no trailing slash). This is considered very useful
+when using templates to do the following thing:
+
+    <link rel="stylesheet" href="[% request.uri_base %]/css/style.css" />
+
+=head2 uri_for(path, params)
+
+Constructs a URI from the base and the passed path. If params (hashref) is
+supplied, these are added to the query string of the URI.
+
+Thus, with the following base:
+
+    http://localhost:5000/foo
+
+You get the following behavior:
+
+    my $uri = request->uri_for('/bar', { baz => 'baz' });
+    print $uri; # http://localhost:5000/foo/bar?baz=baz
+
+C<uri_for> returns a L<URI> object (which can stringify to the value).
+
+=head2 user
+
+Return remote user if defined.
+
 =head2 var
 
 By-name interface to variables stored in this request object.
@@ -724,199 +963,98 @@ returns the value of 'some_variable', while
 
 will set it.
 
-=head2 id()
+=head2 vars
 
-The ID of the request. This allows you to trace a specific request in loggers,
-per the string created using C<to_string>.
+Access to the internal hash of variables:
 
-The ID of the request is essentially the number of requests run in the current
-class.
+    my $value = $request->vars->{'my_key'};
 
-=head2 path()
+You want to use C<var> above.
 
-Return the path requested by the client.
+=head1 Common HTTP request headers
 
-=head2 method()
+Commonly used client-supplied HTTP request headers are available through
+specific accessors:
 
-Return the HTTP method used by the client to access the application.
+=over 4
 
-While this method returns the method string as provided by the environment, it's
-better to use one of the following boolean accessors if you want to inspect the
-requested method.
+=item C<accept>
 
-=head2 content_type()
+HTTP header: C<HTTP_ACCEPT>.
 
-Return the content type of the request.
+=item C<accept_charset>
 
-=head2 content_length()
+HTTP header: C<HTTP_ACCEPT_CHARSET>.
 
-Return the content length of the request.
+=item C<accept_encoding>
 
-=head2 content_encoding()
+HTTP header: C<HTTP_ACCEPT_ENCODING>.
 
-Return the content encoding of the request.
+=item C<accept_language>
 
-=head2 body()
+HTTP header: C<HTTP_ACCEPT_LANGUAGE>.
 
-Return the raw body of the request, unparsed.
+=item C<agent>
 
-If you need to access the body of the request, you have to use this accessor and
-should not try to read C<psgi.input> by hand. C<Dancer2::Core::Request>
-already did it for you and kept the raw body untouched in there.
+Alias for C<user_agent>) below.
 
-=head2 uploads()
+=item C<connection>
 
-Returns a reference to a hash containing uploads. Values can be either a
-L<Dancer2::Core::Request::Upload> object, or an arrayref of
-L<Dancer2::Core::Request::Upload>
-objects.
+HTTP header: C<HTTP_CONNECTION>.
 
-You should probably use the C<upload($name)> accessor instead of manually accessing the
-C<uploads> hash table.
+=item C<content_encoding>
 
-=head2 header($name)
+HTTP header: C<HTTP_CONTENT_ENCODING>.
 
-Return the value of the given header, if present. If the header has multiple
-values, returns an the list of values if called in list context, the first one
-in scalar.
+=item C<content_length>
 
-=head2 new()
+HTTP header: C<HTTP_CONTENT_LENGTH>.
 
-The constructor of the class, used internally by Dancer2's core to create request
-objects.
+=item C<content_type>
 
-It uses the environment hash table given to build the request object:
+HTTP header: C<HTTP_CONTENT_TYPE>.
 
-    Dancer2::Core::Request->new(env => \%env);
+=item C<forwarded_for_address>
 
-It also accepts the C<body_is_parsed> boolean flag, if the new request object should
-not parse request body.
+HTTP header: C<HTTP_X_FORWARDED_FOR>.
 
-=head2 address()
+=item C<forwarded_host>
 
-Return the IP address of the client.
+HTTP header: C<HTTP_X_FORWARDED_HOST>.
 
-=head2 remote_host()
+=item C<forwarded_protocol>
 
-Return the remote host of the client. This only works with web servers configured
-to do a reverse DNS lookup on the client's IP address.
+One of either C<HTTP_X_FORWARDED_PROTOCOL>, C<HTTP_X_FORWARDED_PROTO>, or
+C<HTTP_FORWARDED_PROTO>.
 
-=head2 protocol()
+=item C<host>
 
-Return the protocol (HTTP/1.0 or HTTP/1.1) used for the request.
+Checks whether we are behind a proxy using the C<is_behind_proxy>
+configuration option, and if so returns the first
+C<HTTP_X_FORWARDED_HOST>, since this is a comma separated list.
 
-=head2 port()
+If you have not configured that you behind a proxy, it returns HTTP
+header C<HTTP_HOST>.
 
-Return the port of the server.
+=item C<keep_alive>
 
-=head2 request_uri()
+HTTP header: C<HTTP_KEEP_ALIVE>.
 
-Return the raw, undecoded request URI path.
+=item C<referer>
 
-=head2 user()
+HTTP header: C<HTTP_REFERER>.
 
-Return remote user if defined.
+=item C<user_agent>
 
-=head2 script_name()
+HTTP header: C<HTTP_USER_AGENT>.
 
-Return script_name from the environment.
+=item C<x_requested_with>
 
-=head2 scheme()
+HTTP header: C<HTTP_X_REQUESTED_WITH>.
 
-Return the scheme of the request
+=back
 
-=head2 serializer()
-
-Returns the optional serializer object used to deserialize request parameters.
-
-=head2 data()
-
-If the application has a serializer and if the request has serialized
-content, returns the deserialized structure as a hashref.
-
-=head2 secure()
-
-Return true of false, indicating whether the connection is secure
-
-=head2 uri()
-
-An alias to request_uri()
-
-=head2 is_get()
-
-Return true if the method requested by the client is 'GET'
-
-=head2 is_head()
-
-Return true if the method requested by the client is 'HEAD'
-
-=head2 is_post()
-
-Return true if the method requested by the client is 'POST'
-
-=head2 is_put()
-
-Return true if the method requested by the client is 'PUT'
-
-=head2 is_delete()
-
-Return true if the method requested by the client is 'DELETE'
-
-=head2 request_method
-
-Alias to the C<method> accessor, for backward-compatibility with C<CGI> interface.
-
-=head2 input_handle
-
-Alias to the PSGI input handle (C<< <request->env->{psgi.input}> >>)
-
-=head2 to_string()
-
-Return a string representing the request object (eg: C<"GET /some/path">)
-
-=head2 base()
-
-Returns an absolute URI for the base of the application.  Returns a L<URI>
-object (which stringifies to the URL, as you'd expect).
-
-=head2 uri_base()
-
-Same thing as C<base> above, except it removes the last trailing slash in the
-path if it is the only path.
-
-This means that if your base is I<http://myserver/>, C<uri_base> will return
-I<http://myserver> (notice no trailing slash). This is considered very useful
-when using templates to do the following thing:
-
-    <link rel="stylesheet" href="[% request.uri_base %]/css/style.css" />
-
-=head2 dispatch_path()
-
-The part of the C<path> after C<base>. This is the path used
-for dispatching the request to routes.
-
-=head2 uri_for(path, params)
-
-Constructs a URI from the base and the passed path.  If params (hashref) is
-supplied, these are added to the query string of the uri.  If the base is
-C<http://localhost:5000/foo>, C<< request->uri_for('/bar', { baz => 'baz' }) >>
-would return C<http://localhost:5000/foo/bar?baz=baz>.  Returns a L<URI> object
-(which stringifies to the URL, as you'd expect).
-
-=head2 params($source)
-
-Called in scalar context, returns a hashref of params, either from the specified
-source (see below for more info on that) or merging all sources.
-
-So, you can use, for instance:
-
-    my $foo = params->{foo}
-
-If called in list context, returns a list of key => value pairs, so you could use:
-
-    my %allparams = params;
-
-=head3 Fetching only params from a given source
+=head1 Fetching only params from a given source
 
 If a required source isn't specified, a mixed hashref (or list of key value
 pairs, in list context) will be returned; this will contain params from all
@@ -943,92 +1081,19 @@ returned.
 
 If another value is given for C<$source>, then an exception is triggered.
 
-=head2 is_ajax()
+=head1 EXTRA SPEED
 
-Return true if the value of the header C<X-Requested-With> is XMLHttpRequest.
-
-=head2 upload($name)
-
-Context-aware accessor for uploads. It's a wrapper around an access to the hash
-table provided by C<uploads()>. It looks at the calling context and returns a
-corresponding value.
-
-If you have many file uploads under the same name, and call C<upload('name')> in
-an array context, the accessor will unroll the ARRAY ref for you:
-
-    my @uploads = request->upload('many_uploads'); # OK
-
-Whereas with a manual access to the hash table, you'll end up with one element
-in @uploads, being the ARRAY ref:
-
-    my @uploads = request->uploads->{'many_uploads'}; # $uploads[0]: ARRAY(0xXXXXX)
-
-That is why this accessor should be used instead of a manual access to
-C<uploads>.
-
-=head2 cookies()
-
-Returns a reference to a hash containing cookies, where the keys are the names of the
-cookies and values are L<Dancer2::Core::Cookie> objects.
-
-=head1 Common HTTP request headers
-
-Commonly used client-supplied HTTP request headers are available through
-specific accessors, here are those supported:
+If L<Dancer2::Core::Request> detects the following modules as installed,
+it will use them to speed things up:
 
 =over 4
 
-=item C<accept>
+=item * L<URL::Encode::XS>
 
-=item C<accept_charset>
-
-=item C<accept_encoding>
-
-=item C<accept_language>
-
-=item C<agent> (alias for C<user_agent>)
-
-=item C<connection>
-
-=item C<forwarded_for_address>
-
-=item C<forwarded_protocol>
-
-=item C<forwarded_host>
-
-=item C<host>
-
-=item C<keep_alive>
-
-=item C<path_info>
-
-=item C<referer>
-
-=item C<remote_address>
-
-=item C<user_agent>
-
-=item C<x_requested_with>
+=item * L<CGI::Deurl::XS>
 
 =back
 
-With the exception of C<host>, these accessors are lookups into the PSGI env
-hash reference.
-
-Note that the L<PSGI> specification prefixes client-supplied request headers with
-C<HTTP_>. For example, a C<X-Requested-With> header has the key
-C<HTTP_X_REQUESTED_WITH> in the PSGI env hashref.
-
-=head1 EXTRA SPEED
-
-Install URL::Encode::XS and CGI::Deurl::XS for extra speed.
-
-Dancer2::Core::Request will use it if they detect their presence.
-
-=head1 SEE ALSO
-
-L<Dancer2>
-
 =head1 AUTHOR
 
 Dancer Core Developers
